]> git.cameronkatri.com Git - mandoc.git/blobdiff - html.c
simplify a few accesses to fields of structs, using auxiliary pointer
[mandoc.git] / html.c
diff --git a/html.c b/html.c
index 15b87236e5f4b24f18f9044432c0d68cf5df1c3d..71c9c711e825851b2b52bec5a6d58d5904dcde11 100644 (file)
--- a/html.c
+++ b/html.c
@@ -1,7 +1,7 @@
-/* $Id: html.c,v 1.266 2020/04/07 22:56:02 schwarze Exp $ */
+/* $Id: html.c,v 1.275 2021/09/09 14:47:24 schwarze Exp $ */
 /*
- * Copyright (c) 2011-2015, 2017-2020 Ingo Schwarze <schwarze@openbsd.org>
  * Copyright (c) 2008-2011, 2014 Kristaps Dzonsons <kristaps@bsd.lv>
+ * Copyright (c) 2011-2015, 2017-2021 Ingo Schwarze <schwarze@openbsd.org>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -81,7 +81,7 @@ static        const struct htmldata htmltags[TAG_MAX] = {
        {"h1",          HTML_TOPHRASE | HTML_NLAROUND},
        {"h2",          HTML_TOPHRASE | HTML_NLAROUND},
        {"p",           HTML_TOPHRASE | HTML_NLAROUND | HTML_INDENT},
-       {"pre",         HTML_TOPHRASE | HTML_NLALL | HTML_NOINDENT},
+       {"pre",         HTML_TOPHRASE | HTML_NLAROUND | HTML_NOINDENT},
        {"a",           HTML_INPHRASE | HTML_TOPHRASE},
        {"b",           HTML_INPHRASE | HTML_TOPHRASE},
        {"cite",        HTML_INPHRASE | HTML_TOPHRASE},
@@ -91,6 +91,7 @@ static        const struct htmldata htmltags[TAG_MAX] = {
        {"span",        HTML_INPHRASE | HTML_TOPHRASE},
        {"var",         HTML_INPHRASE | HTML_TOPHRASE},
        {"br",          HTML_INPHRASE | HTML_NOSTACK | HTML_NLALL},
+       {"hr",          HTML_INPHRASE | HTML_NOSTACK},
        {"mark",        HTML_INPHRASE },
        {"math",        HTML_INPHRASE | HTML_NLALL | HTML_INDENT},
        {"mrow",        0},
@@ -112,6 +113,11 @@ static     const struct htmldata htmltags[TAG_MAX] = {
 };
 
 /* Avoid duplicate HTML id= attributes. */
+
+struct id_entry {
+       int      ord;   /* Ordinal number of the latest occurrence. */
+       char     id[];  /* The id= attribute without any ordinal suffix. */
+};
 static struct ohash     id_unique;
 
 static void     html_reset_internal(struct html *);
@@ -135,6 +141,7 @@ html_alloc(const struct manoutput *outopts)
        h = mandoc_calloc(1, sizeof(struct html));
 
        h->tag = NULL;
+       h->metac = h->metal = ESCAPE_FONTROMAN;
        h->style = outopts->style;
        if ((h->base_man1 = outopts->man) == NULL)
                h->base_man2 = NULL;
@@ -146,7 +153,7 @@ html_alloc(const struct manoutput *outopts)
        if (outopts->toc)
                h->oflags |= HTML_TOC;
 
-       mandoc_ohash_init(&id_unique, 4, 0);
+       mandoc_ohash_init(&id_unique, 4, offsetof(struct id_entry, id));
 
        return h;
 }
@@ -155,17 +162,17 @@ static void
 html_reset_internal(struct html *h)
 {
        struct tag      *tag;
-       char            *cp;
+       struct id_entry *entry;
        unsigned int     slot;
 
        while ((tag = h->tag) != NULL) {
                h->tag = tag->next;
                free(tag);
        }
-       cp = ohash_first(&id_unique, &slot);
-       while (cp != NULL) {
-               free(cp);
-               cp = ohash_next(&id_unique, &slot);
+       entry = ohash_first(&id_unique, &slot);
+       while (entry != NULL) {
+               free(entry);
+               entry = ohash_next(&id_unique, &slot);
        }
        ohash_delete(&id_unique);
 }
@@ -174,7 +181,7 @@ void
 html_reset(void *p)
 {
        html_reset_internal(p);
-       mandoc_ohash_init(&id_unique, 4, 0);
+       mandoc_ohash_init(&id_unique, 4, offsetof(struct id_entry, id));
 }
 
 void
@@ -190,6 +197,8 @@ print_gen_head(struct html *h)
        struct tag      *t;
 
        print_otag(h, TAG_META, "?", "charset", "utf-8");
+       print_otag(h, TAG_META, "??", "name", "viewport",
+           "content", "width=device-width, initial-scale=1.0");
        if (h->style != NULL) {
                print_otag(h, TAG_LINK, "?h??", "rel", "stylesheet",
                    h->style, "type", "text/css", "media", "all");
@@ -232,8 +241,10 @@ html_setfont(struct html *h, enum mandoc_esc font)
        case ESCAPE_FONTITALIC:
        case ESCAPE_FONTBOLD:
        case ESCAPE_FONTBI:
-       case ESCAPE_FONTCW:
        case ESCAPE_FONTROMAN:
+       case ESCAPE_FONTCR:
+       case ESCAPE_FONTCB:
+       case ESCAPE_FONTCI:
                break;
        case ESCAPE_FONT:
                font = ESCAPE_FONTROMAN;
@@ -264,9 +275,17 @@ print_metaf(struct html *h)
                h->metaf = print_otag(h, TAG_B, "");
                print_otag(h, TAG_I, "");
                break;
-       case ESCAPE_FONTCW:
+       case ESCAPE_FONTCR:
                h->metaf = print_otag(h, TAG_SPAN, "c", "Li");
                break;
+       case ESCAPE_FONTCB:
+               h->metaf = print_otag(h, TAG_SPAN, "c", "Li");
+               print_otag(h, TAG_B, "");
+               break;
+       case ESCAPE_FONTCI:
+               h->metaf = print_otag(h, TAG_SPAN, "c", "Li");
+               print_otag(h, TAG_I, "");
+               break;
        default:
                break;
        }
@@ -329,23 +348,27 @@ html_fillmode(struct html *h, enum roff_tok want)
  * element and/or as a segment identifier for a URI in an <a> element.
  * The function may fail and return NULL if the node lacks text data
  * to create the attribute from.
- * If the "unique" argument is 0, the caller is responsible for
- * free(3)ing the returned string after using it.
+ * The caller is responsible for free(3)ing the returned string.
+ *
  * If the "unique" argument is non-zero, the "id_unique" ohash table
- * is used for de-duplication and owns the returned string, so the
- * caller must not free(3) it.  In this case, it will be freed
- * automatically by html_reset() or html_free().
+ * is used for de-duplication.  If the "unique" argument is 1,
+ * it is the first time the function is called for this tag and
+ * location, so if an ordinal suffix is needed, it is incremented.
+ * If the "unique" argument is 2, it is the second time the function
+ * is called for this tag and location, so the ordinal suffix
+ * remains unchanged.
  */
 char *
 html_make_id(const struct roff_node *n, int unique)
 {
        const struct roff_node  *nch;
-       char                    *buf, *bufs, *cp;
+       struct id_entry         *entry;
+       char                    *buf, *cp;
+       size_t                   len;
        unsigned int             slot;
-       int                      suffix;
 
-       if (n->string != NULL)
-               buf = mandoc_strdup(n->string);
+       if (n->tag != NULL)
+               buf = mandoc_strdup(n->tag);
        else {
                switch (n->tok) {
                case MDOC_Sh:
@@ -374,11 +397,12 @@ html_make_id(const struct roff_node *n, int unique)
         * permitted in URL-fragment strings according to the
         * explicit list at:
         * https://url.spec.whatwg.org/#url-fragment-string
+        * In addition, reserve '~' for ordinal suffixes.
         */
 
        for (cp = buf; *cp != '\0'; cp++)
                if (isalnum((unsigned char)*cp) == 0 &&
-                   strchr("!$&'()*+,-./:;=?@_~", *cp) == NULL)
+                   strchr("!$&'()*+,-./:;=?@_", *cp) == NULL)
                        *cp = '_';
 
        if (unique == 0)
@@ -386,25 +410,21 @@ html_make_id(const struct roff_node *n, int unique)
 
        /* Avoid duplicate HTML id= attributes. */
 
-       bufs = NULL;
-       suffix = 1;
        slot = ohash_qlookup(&id_unique, buf);
-       cp = ohash_find(&id_unique, slot);
-       if (cp != NULL) {
-               while (cp != NULL) {
-                       free(bufs);
-                       if (++suffix > 127) {
-                               free(buf);
-                               return NULL;
-                       }
-                       mandoc_asprintf(&bufs, "%s_%d", buf, suffix);
-                       slot = ohash_qlookup(&id_unique, bufs);
-                       cp = ohash_find(&id_unique, slot);
-               }
-               free(buf);
-               buf = bufs;
+       if ((entry = ohash_find(&id_unique, slot)) == NULL) {
+               len = strlen(buf) + 1;
+               entry = mandoc_malloc(sizeof(*entry) + len);
+               entry->ord = 1;
+               memcpy(entry->id, buf, len);
+               ohash_insert(&id_unique, slot, entry);
+       } else if (unique == 1)
+               entry->ord++;
+
+       if (entry->ord > 1) {
+               cp = buf;
+               mandoc_asprintf(&buf, "%s~%d", cp, entry->ord);
+               free(cp);
        }
-       ohash_insert(&id_unique, slot, buf);
        return buf;
 }
 
@@ -494,8 +514,10 @@ print_encode(struct html *h, const char *p, const char *pend, int norecurse)
                case ESCAPE_FONTBOLD:
                case ESCAPE_FONTITALIC:
                case ESCAPE_FONTBI:
-               case ESCAPE_FONTCW:
                case ESCAPE_FONTROMAN:
+               case ESCAPE_FONTCR:
+               case ESCAPE_FONTCB:
+               case ESCAPE_FONTCI:
                        if (0 == norecurse) {
                                h->flags |= HTML_NOSPACE;
                                if (html_setfont(h, esc))
@@ -786,7 +808,7 @@ print_otag_id(struct html *h, enum htmltag elemtype, const char *cattr,
        if (n->flags & NODE_ID)
                id = html_make_id(n, 1);
        if (n->flags & NODE_HREF)
-               href = id == NULL ? html_make_id(n, 0) : id;
+               href = id == NULL ? html_make_id(n, 2) : id;
        if (href != NULL && htmltags[elemtype].flags & HTML_INPHRASE)
                ret = print_otag(h, TAG_A, "chR", "permalink", href);
        t = print_otag(h, elemtype, "ci", cattr, id);
@@ -804,6 +826,7 @@ print_otag_id(struct html *h, enum htmltag elemtype, const char *cattr,
                                print_otag(h, TAG_A, "chR", "permalink", href);
                }
        }
+       free(id);
        if (id == NULL)
                free(href);
        return ret;
@@ -878,6 +901,15 @@ print_gen_comment(struct html *h, struct roff_node *n)
 void
 print_text(struct html *h, const char *word)
 {
+       print_tagged_text(h, word, NULL);
+}
+
+void
+print_tagged_text(struct html *h, const char *word, struct roff_node *n)
+{
+       struct tag      *t;
+       char            *href;
+
        /*
         * Always wrap text in a paragraph unless already contained in
         * some flow container; never put it directly into a section.
@@ -898,13 +930,20 @@ print_text(struct html *h, const char *word)
        }
 
        /*
* Print the text, optionally surrounded by HTML whitespace,
* optionally manually switching fonts before and after.
         */
 
        assert(h->metaf == NULL);
        print_metaf(h);
        print_indent(h);
+
+       if (n != NULL && (href = html_make_id(n, 2)) != NULL) {
+               t = print_otag(h, TAG_A, "chR", "permalink", href);
+               free(href);
+       } else
+               t = NULL;
+
        if ( ! print_encode(h, word, NULL, 0)) {
                if ( ! (h->flags & HTML_NONOSPACE))
                        h->flags &= ~HTML_NOSPACE;
@@ -915,7 +954,8 @@ print_text(struct html *h, const char *word)
        if (h->metaf != NULL) {
                print_tagq(h, h->metaf);
                h->metaf = NULL;
-       }
+       } else if (t != NULL)
+               print_tagq(h, t);
 
        h->flags &= ~HTML_IGNDELIM;
 }
@@ -898,13 +930,20 @@ print_text(struct html *h, const char *word)
 
        /*
-        * Print the text, optionally surrounded by HTML whitespace,
-        * optionally manually switching fonts before and after.
+        * Optionally switch fonts, optionally write a permalink, then
+        * print the text, optionally surrounded by HTML whitespace.
         */
 